# coding=utf-8
import heapq, random
from collections import deque

class BPQueue(object):
    def __init__(self, limit=None, *args):
        self.limit = limit
        self.queue = list()

    def __getitem__(self, val):
        return self.queue[val]

    def __len__(self):
        return len(self.queue)

    def append(self, x):
        #Push the value item onto the heap, maintaining the heap invariant.
        heapq.heappush(self.queue, x)  
        if self.limit and len(self.queue) > self.limit:
            self.queue.remove(heapq.nlargest(1, self.queue)[0])

    def extend(self, iterable):
        for x in iterable:
            self.append(x)



from utils import BPQueue
from models import SearchNodeValueOrdered
import math
import random

def allExpander(fringe, iteration, viewer):
    '''
    Expand all nodes on the fringe.
    '''
    expanded_neighbors = [node.expand(local_search=True)
                          for node in fringe]
    list(map(fringe.extend, expanded_neighbors))

def firstExpander(fringe, iteration):
    '''
    Expands only the first node on the fringe.
    '''
    curr = fringe[0]
    neighbors = curr.expand(local_search=True)
    fringe.extend(neighbors)


def hillClimbing(problem, iterationsLimit=0):
    '''
    Hill Climbing search algorithm.

    If iterationsLimit is specified, the algorithm will end after that
    number of iterations. Else, it will continue until it can't find a
    better node than the current one.
    This algorithm requires: SearchProblem.actions, SearchProblem.result, and
    SearchProblem.value.
    '''
    return localSearchAlg(problem, firstExpander,
                         iterationsLimit=iterationsLimit,
                         fringe_size=1, stop_when_no_better=True)

def localSearchAlg(problem, fringe_expander, iterationsLimit=0, fringe_size=1,
                  random_initial_states=False, stop_when_no_better=True):
    '''
    Basic Algorithm for all Local Search algorithms.
    This algorithm requires: 
    SearchProblem.actions, SearchProblem.result, SearchProblem.value.
    '''
    # work with queue - append , pop, extend, clear, remove or sorted
    fringe = BPQueue(fringe_size)
    #in case of random values
    fringe.append(SearchNodeValueOrdered(state=problem.initial_state, problem=problem))

    iteration = 0
    run = True
    best = None

    while run:

        old_best = fringe[0]
        fringe_expander(fringe, iteration)
        best = fringe[0]
        iteration += 1

        if iterationsLimit and iteration >= iterationsLimit:
            run = False

        elif old_best.value >= best.value and stop_when_no_better:
            run = False
            
    return best



class SearchProblem(object):
    '''Abstract class to manipulate the search space of the problem, and it will be
       represented as a graph. Each state corresponds with a problem state and
       each problem action corresponds with an edge.
    '''

    def __init__(self, initial_state=None):
        self.initial_state = initial_state

    def actions(self, state):
        '''Returns the actions available to perform from `state`.
           The returned value is an iterable over actions.
           Actions are problem-specific and no assumption should be made about
           them.
        '''
        raise NotImplementedError

    def result(self, state, action):
        '''Returns the resulting state of applying `action` to `state`.'''
        raise NotImplementedError

    def cost(self, state, action, state2):
        '''Returns the cost of applying `action` from `state` to `state2`.
           The returned value is a number (integer or floating point).
           By default this function returns `1`.
        '''
        return 1

    def value(self, state):
        '''Returns the value of `state` and it's a number (int or float).'''
        raise NotImplementedError

class SearchNode(object):
    '''Node for the process of searching.'''

    def __init__(self, state, parent=None, action=None, cost=0, problem=None, depth=0):
        self.state = state
        self.parent = parent
        self.action = action
        self.cost = cost
        self.problem = problem or parent.problem
        self.depth = depth

    def expand(self, local_search=False):
        '''Create successors.'''
        new_nodes = [] #list
        for action in self.problem.actions(self.state):
            new_state = self.problem.result(self.state, action)
            #print(new_state) # all steps untill the end
            cost = self.problem.cost(self.state, action, new_state)
            ''' Reference to the type of the current instance. '''
            nodefactory = self.__class__
            new_nodes.append(nodefactory(state=new_state, parent=None if local_search else self,
                                         problem=self.problem, action=action,
                                         cost=self.cost + cost, depth=self.depth + 1))
        return new_nodes

    def path(self):
        '''Path (list of nodes and actions) from root to this node.'''
        node = self
        path = []
        while node:
            path.append((node.action, node.state))
            node = node.parent
        return list(reversed(path))

class SearchNodeValueOrdered(SearchNode):
    def __init__(self, *args, **kwargs):
        super(SearchNodeValueOrdered, self).__init__(*args, **kwargs)
        ''' Check the values'''
        self.value = self.problem.value(self.state)

    def __lt__(self, other):
        #is equivalent to self,other
        # value must work inverted, because heapq sorts 1-9
        # and we need 9-1 sorting
        return -self.value < -other.value


#!/usr/bin/env python
# -*- coding: utf-8 -*-
import setuptools
from distutils.core import setup

setup(
    name='LSpackage',
    version='1.0',
    description=u'An implementation of Local Search algorithms based on single solution design',
    long_description=open('README.rst').read(),
    author=u'Andurnache Alexandru',
    author_email='alex.andur@yahoo.com',
    url='http://github.com/alexandur/LSpackage',
    packages=['code'],
    license='LICENSE.txt',
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
    ],
)
